import grpc
import logging

from perxis.extensions import extension_pb2, extension_pb2_grpc
from perxis.roles import roles_pb2_grpc, roles_pb2
from perxis.clients import clients_pb2_grpc, clients_pb2
from perxis.collections import collections_pb2_grpc, collections_pb2
from perxis.environments import environments_pb2_grpc
from perxis.extensions.extension_setup import ExtensionSetup


class ExtensionService(extension_pb2_grpc.ExtensionServicer):
    extension_id: str
    collections: list[collections_pb2.Collection] = []
    roles: list[roles_pb2.Role] = []
    clients: list[clients_pb2.Client] = []

    def __init__(self,
                 collections_service: collections_pb2_grpc.CollectionsStub,
                 environments_service: environments_pb2_grpc.EnvironmentsStub,
                 roles_service: roles_pb2_grpc.RolesStub,
                 clients_service: clients_pb2_grpc.ClientsStub,
    ):
        self.logger = logging.getLogger(__name__)
        self.collections_service = collections_service
        self.environments_service = environments_service
        self.roles_service = roles_service
        self.clients_service = clients_service

        self.extension_setup = ExtensionSetup(
            self.collections_service, self.environments_service,
            self.roles_service, self.clients_service
        )

        for collection in self.collections or []:
            self.extension_setup.add_collection(collection)

        for role in self.roles or []:
            self.extension_setup.add_role(role)

        for client in self.clients or []:
            self.extension_setup.add_client(client)

    def write_result_log(self, operation, request, response):
        log_func = self.get_log_func_by_state(response.results[0].state)

        log_func(
            "Результат %s расширения %s для окружения %s пространства %s: \r\n"
            "State: %s\r\n"
            "Msg: %s\r\n"
            "Error: %s" % (
                operation,
                self.extension_id,
                request.env_id,
                request.space_id,
                response.results[0].state,
                response.results[0].msg,
                response.results[0].error
            )
        )

    def get_log_func_by_state(self, state):
        if state == extension_pb2.ExtensionRequestResult.State.OK:
            return self.logger.info
        else:
            return self.logger.error

    def ext_request_results_from_exception(self, e: Exception) -> extension_pb2.ExtensionRequestResult:
        return [
            extension_pb2.ExtensionRequestResult(
                extension=self.extension_id,
                state=extension_pb2.ExtensionRequestResult.State.ERROR,
                error=str(e),
                msg=None
            )
        ]

    async def _Install(self, request: extension_pb2.InstallRequest, context):
        errors_list = await self.extension_setup.install(
            request.space_id, request.env_id, request.force
        )

        response_state = extension_pb2.ExtensionRequestResult.State.OK \
            if not errors_list \
            else extension_pb2.ExtensionRequestResult.State.ERROR

        return extension_pb2.InstallResponse(
            results=[extension_pb2.ExtensionRequestResult(
                extension=self.extension_id,
                state=response_state,
                error="; ".join(errors_list) if errors_list else None,
                msg="Ok" if not errors_list else None
            )]
        )

    async def Install(self, request: extension_pb2.InstallRequest, context):
        self.logger.info(
            "Установка расширения %s для окружения %s пространства %s. %s force" % (
                self.extension_id,
                request.env_id,
                request.space_id,
                "С" if request.force else "Без"
            )
        )

        try:
            response = await self._Install(request, context)
        except Exception as e:
            response = extension_pb2.InstallResponse(
                results=self.ext_request_results_from_exception(e)
            )

        if response.results[0].state != extension_pb2.ExtensionRequestResult.State.OK:
            context.set_code(grpc.StatusCode.UNKNOWN)
            context.set_details(response.results[0].error)

        self.write_result_log("установки", request, response)

        return response

    async def _Update(self, request: extension_pb2.UpdateRequest, context):
        errors_list = await self.extension_setup.update(
            request.space_id, request.env_id, request.force
        )

        response_state = extension_pb2.ExtensionRequestResult.State.OK \
            if not errors_list \
            else extension_pb2.ExtensionRequestResult.State.ERROR

        return extension_pb2.UpdateResponse(
            results=[extension_pb2.ExtensionRequestResult(
                extension=self.extension_id,
                state=response_state,
                error="; ".join(errors_list) if errors_list else None,
                msg="Ok" if not errors_list else None
            )]
        )

    async def Update(self, request: extension_pb2.UpdateRequest, context):
        self.logger.info(
            "Обновление расширения %s для окружения %s пространства %s. %s force" % (
                self.extension_id,
                request.env_id,
                request.space_id,
                "С" if request.force else "Без"
            )
        )

        try:
            response = await self._Update(request, context)
        except Exception as e:
            response = extension_pb2.UpdateResponse(
                results=self.ext_request_results_from_exception(e)
            )

        if response.results[0].state != extension_pb2.ExtensionRequestResult.State.OK:
            context.set_code(grpc.StatusCode.UNKNOWN)
            context.set_details(response.results[0].error)

        self.write_result_log("обновления", request, response)

        return response

    async def _Uninstall(self, request: extension_pb2.UninstallRequest, context):
        errors_list: list[str] = await self.extension_setup.uninstall(request.space_id, request.env_id, request.remove)

        response_state = extension_pb2.ExtensionRequestResult.State.OK \
            if not errors_list \
            else extension_pb2.ExtensionRequestResult.State.ERROR

        return extension_pb2.UninstallResponse(
            results=[extension_pb2.ExtensionRequestResult(
                extension=self.extension_id,
                state=response_state,
                error="; ".join(errors_list) if errors_list else None,
                msg="Ok" if not errors_list else None
            )]
        )

    async def Uninstall(self, request: extension_pb2.UninstallRequest, context):
        self.logger.info(
            "Удаление расширения %s для окружения %s пространства %s. %s remove" % (
                self.extension_id,
                request.env_id,
                request.space_id,
                "С" if request.remove else "Без"
            )
        )

        try:
            response = await self._Uninstall(request, context)
        except Exception as e:
            response = extension_pb2.UninstallResponse(
                results=self.ext_request_results_from_exception(e)
            )

        if response.results[0].state != extension_pb2.ExtensionRequestResult.State.OK:
            context.set_code(grpc.StatusCode.UNKNOWN)
            context.set_details(response.results[0].error)

        self.write_result_log("удаления", request, response)

        return response

    async def _Check(self, request: extension_pb2.CheckRequest, context):
        errors_list = await self.extension_setup.check(request.space_id, request.env_id)

        response_state = extension_pb2.ExtensionRequestResult.State.OK \
            if not errors_list \
            else extension_pb2.ExtensionRequestResult.State.ERROR

        return extension_pb2.CheckResponse(
            results=[extension_pb2.ExtensionRequestResult(
                extension=self.extension_id,
                state=response_state,
                error="; ".join(errors_list) if errors_list else None,
                msg="Ok" if not errors_list else None
            )]
        )

    async def Check(self, request: extension_pb2.CheckRequest, context):
        self.logger.info(
            "Проверка расширения %s для окружения %s пространства %s" % (
                self.extension_id,
                request.env_id,
                request.space_id,
            )
        )

        try:
            response = await self._Check(request, context)
        except Exception as e:
            response = extension_pb2.CheckResponse(
                results=self.ext_request_results_from_exception(e)
            )

        if response.results[0].state != extension_pb2.ExtensionRequestResult.State.OK:
            context.set_code(grpc.StatusCode.UNKNOWN)
            context.set_details(response.results[0].error)

        self.write_result_log("проверки", request, response)

        return response

    async def Action(self, request: extension_pb2.ActionRequest, context):
        context.set_code(grpc.StatusCode.UNKNOWN)
        context.set_details("Unknown action")

        return None
